## TODO move this logic to another repo.
# from atlasvar.probes.models import Mutation
# from atlasvar.probes import AlleleGenerator
# from atlasvar.probes import make_variant_probe
# from atlasvar.annotation.genes import GeneAminoAcidChangeToDNAVariants


# import logging
# logging.basicConfig()
# logger = logging.getLogger(__name__)


# def flatten(l): return [item for sublist in l for item in sublist]


# class BIGSIVariantSearch(object):

#     def __init__(self, bigsi, reference):
#         self.bigsi = bigsi
#         self.reference = reference
#         self.al = AlleleGenerator(reference_filepath=self.reference,
#                                   kmer=self.bigsi.kmer_size)

#     def search_for_alleles(self, ref_seqs, alt_seqs):
#         results = {"ref": [], "alt": []}
#         for ref in ref_seqs:
#             res = self.bigsi.search(ref, score=False)
#             results["ref"].extend(res.keys())
#         for alt in alt_seqs:
#             res = self.bigsi.search(alt, score=False)
#             results["alt"].extend(res.keys())
#         return results

#     def make_variant_probe_set(self, var):
#         return make_variant_probe(
#             self.al, var, self.bigsi.kmer_size, DB=None)

#     def create_variant_probe_set(self, var_name):
#         var = Mutation(var_name=var_name, reference=self.reference).variant
#         variant_panel = self.make_variant_probe_set(var)
#         return variant_panel

#     def search_for_variant(self, ref_base, pos, alt_base="X", alphabet="DNA"):
#         if not alphabet in ["DNA", "PROT"]:
#             raise ValueError("alphabet must be either DNA or PROT")
#         var_name = "".join([ref_base, str(pos), alt_base])
#         variant_probe_set = self.create_variant_probe_set(var_name=var_name)
#         return {var_name: self.genotype_alleles(variant_probe_set.refs, variant_probe_set.alts)}

#     def genotype_alleles(self, refs, alts):
#         ref_alt_samples = self.search_for_alleles(refs, alts)
#         results = {}
#         for sample_id in set(flatten(ref_alt_samples.values())):
#             if sample_id in ref_alt_samples["ref"] and sample_id in ref_alt_samples["alt"]:
#                 results[sample_id] = {"genotype": "0/1"}
#             elif sample_id in ref_alt_samples["ref"]:
#                 results[sample_id] = {"genotype": "0/0"}
#             elif sample_id in ref_alt_samples["alt"]:
#                 results[sample_id] = {"genotype": "1/1"}
#         return results


# class BIGSIAminoAcidMutationSearch(BIGSIVariantSearch):

#     def __init__(self, bigsi, reference, genbank):
#         super(BIGSIAminoAcidMutationSearch, self).__init__(bigsi, reference)
#         self.genbank = genbank
#         self.aa2dna = GeneAminoAcidChangeToDNAVariants(
#             self.reference,
#             self.genbank)

#     def search_for_amino_acid_variant(self, gene, ref, pos, alt):
#         mut_name = "".join([ref, str(pos), alt])
#         gene_mut_name = "_".join([gene, mut_name])
#         results = {gene_mut_name: {}}
#         _results = results[gene_mut_name]

#         for var_name in self.aa2dna.get_variant_names(gene, mut_name, True):
#             mut = Mutation(reference=self.reference,
#                            var_name=var_name,
#                            gene=gene,
#                            mut=mut_name)
#             variant_probe_set = self.create_variant_probe_set(var_name)
#             variant_calls = self.genotype_alleles(
#                 variant_probe_set.refs, variant_probe_set.alts)
#             for sample, genotype in variant_calls.items():
#                 _results[sample] = {"genotype": genotype[
#                     "genotype"], "aa_mut": mut.mut, "variant": mut.variant.var_name, "gene": gene}
#         return results



from bigsi.storage.base import BaseStorage
from bigsi.utils import batch
from bigsi.constants import DEFAULT_ROCKS_DB_STORAGE_CONFIG
import rocksdb
import shutil
import copy
import gc
import os


class RocksDB(rocksdb.DB):
    def __setitem__(self, key, val):
        self.put(key, val)

    def __getitem__(self, key):
        val = self.get(key)
        if val is None:
            raise KeyError("%s does not exist" % key)
        return val


class RocksDBStorage(BaseStorage):
    def __init__(self, storage_config=None):
        if storage_config is None:
            storage_config = DEFAULT_ROCKS_DB_STORAGE_CONFIG
        self.storage_config = copy.copy(storage_config)
        options = storage_config["options"]

        self.storage = RocksDB(
            self.storage_config["filename"], rocksdb.Options(**options)
        )
        self.write_batch_size = int(self.storage_config.get("write_batch_size", 10000))

    def __repr__(self):
        return "rocksdb storage"

    def delete_all(self):
        try:
            shutil.rmtree(self.storage_config["filename"])
        except FileNotFoundError:
            pass
        del self.storage
        RocksDBStorage.__init__(self, self.storage_config)

    def batch_set(self, keys, values):
        for batchiter in batch(zip(keys, values), self.write_batch_size):
            writebatch = rocksdb.WriteBatch()
            for k, v in batchiter:
                writebatch.put(k, v)
            self.storage.write(writebatch)

    def batch_get(self, keys):
        keys = list(keys)
        result = self.storage.multi_get(keys)
        return [result[k] for k in keys]

    def sync(self):
        gc.collect()

    def close(self):
        self.__delete_lock_file()
        del self.storage
        gc.collect()

    def __delete_lock_file(self):
        lock_file = os.path.join(self.storage_config["filename"], "LOCK")
        try:
            os.remove(lock_file)
        except (FileNotFoundError, NotADirectoryError, PermissionError):
            pass
        gc.collect()



from bitarray import bitarray
import numpy as np


class BitArray(bitarray):
    def __init__(self, *args, **kwargs):
        super().__init__()

    def setbit(self, i, bit):
        if i < 0:
            raise ValueError("Index must be >= 0")
        try:
            self[i] = bit
            return self
        except IndexError:
            self.extend([False] * (1 + i - self.length()))
            return self.setbit(i, bit)

    def getbit(self, i):
        try:
            return self[i]
        except IndexError:
            return False

    def indexes(self):
        return np.where(self)[0].tolist()

    def colours(self):
        return self.indexes()



from bitarray import bitarray
import copy
import logging

logger = logging.getLogger(__name__)
from bigsi.utils import DEFAULT_LOGGING_LEVEL

logger.setLevel(DEFAULT_LOGGING_LEVEL)
import numpy as np

logger = logging.getLogger(__name__)


def transpose_low_mem(bitarrays):
    logger.info("Using slow, low memory transpose")
    # Takes a list of bitarrays and returns the transpose as a list of
    # bitarrays
    x = len(bitarrays)
    y = bitarrays[0].length()
    logger.info("BFM dims %i %i" % (x, y))

    tbitarrays = []
    for ii in range(y):
        ba = bitarray(x)
        ba.setall(False)
        tbitarrays.append(ba)
    for i in range(x):
        for j in range(y):
            tbitarrays[j][i] = bitarrays[i][j]
    return tbitarrays


# def transpose_numpy(bitarrays):
#     logger.info("Using high memory transpose")

#     # Takes a list of bitarrays and returns the transpose as a list of
#     # bitarrays
#     X = np.array(bitarrays).transpose().copy()
#     X=np.array([np.frombuffer(ba.unpack(),dtype=bool) for ba in bitarrays],dtype=bool).transpose().copy()
#     tbitarrays=[]
#     for row in X:
#         ba=bitarray()
#         ba.pack(row.tobytes())
#         tbitarrays.append(ba)
#     # tbitarrays=[bitarray(i.tolist()) for i in X]
#     return tbitarrays


def transpose_numpy(bitarrays):
    # Takes a list of bitarrays and returns the transpose as a list of
    # bitarrays

    X = np.array([np.frombuffer(ba.unpack(), dtype=bool) for ba in bitarrays], dtype=bool)
    for row in X.T:
        ba = bitarray()
        ba.pack(row.tobytes())
        yield ba


def transpose(bitarrays, lowmem=False):
    if lowmem:
        return transpose_low_mem(bitarrays)
    else:
        return transpose_numpy(bitarrays)



from bitarray import bitarray
import numpy as np
import math
import copy


def remove_short_ones(s):
    ba = bitarray(s)
    if len(ba) < 3:
        return ba.to01()
    ba2 = ba[1:]
    ba2.append(1)
    ba3 = ba2[1:]
    ba3.append(1)
    ba_out = ba & ba2 & ba3
    return ba_out.to01()


def tabulate_score(ss):
    score_counter = {"0": [], "1": []}
    cnt = 1
    for i in range(len(ss)):
        current = ss[i]
        if i < len(ss) - 1:
            _next = ss[i + 1]
            cnt += 1
            if current != _next:
                score_counter[current].append(cnt)
                cnt = 1
        else:
            score_counter[current].append(cnt)
    return score_counter


class Scorer:
    def __init__(
        self, DB_SIZE, MATCH=1, MISMATCH=2, LAMBDA_UNGAPPED=1.330, K_UNGAPPED=0.621, LAMBDA_GAPPED=1.28, K_GAPPED=0.46
    ):
        self.LAMBDA_UNGAPPED = LAMBDA_UNGAPPED
        self.K_UNGAPPED = K_UNGAPPED
        self.LAMBDA_GAPPED = LAMBDA_GAPPED
        self.K_GAPPED = K_GAPPED

        self.MATCH = MATCH
        self.DB_SIZE = DB_SIZE
        self.MISMATCH = MISMATCH
        self.kmer_adjust = 3

    def calculate_score(self, score_counter, convert):
        max_score = copy.copy(self.MATCH * sum(score_counter["1"]))
        min_score = copy.copy(max_score)
        mean_score = copy.copy(min_score)

        SNP_t = 31 + self.kmer_adjust  # correct for the 'remove_short_ones'
        max_total_N_snps = 0
        min_total_N_snps = 0
        for i in score_counter["0"]:
            min_N_snps = float(i) / SNP_t
            max_N_snps = (i - SNP_t) + 1
            if max_N_snps < min_N_snps:
                max_N_snps = min_N_snps
            max_total_N_snps += max_N_snps
            min_total_N_snps += min_N_snps
            mean_N_snps = min_N_snps + 0.05 * max_N_snps

            max_penalty = self.MISMATCH * (max_N_snps)
            min_penalty = self.MISMATCH * (min_N_snps)
            mean_penalty = self.MISMATCH * (mean_N_snps)

            points_for_max = self.MATCH * (i - max_penalty)
            points_for_min = self.MATCH * (i - min_penalty)
            points_for_mean = self.MATCH * (i - mean_penalty)

            max_score = round(max_score - min_penalty + points_for_min, 2)
            min_score = round(min_score - max_penalty + points_for_max, 2)
            mean_score = round(mean_score - mean_penalty + points_for_mean, 2)

        return {
            "score": round(mean_score * convert, 2),
            "min_score": round(min_score * convert, 2),
            "max_score": round(max_score * convert, 2),
            "max_mismatches": math.ceil(max_total_N_snps),
            "min_mismatches": math.floor(min_total_N_snps),
            "mismatches": math.ceil(math.ceil(min_total_N_snps) + (0.05 * math.floor(max_total_N_snps))),
        }

    def score(self, s):
        ss = remove_short_ones(s)
        max_possible_score = len(ss)
        seq_len = max_possible_score + 31 - 1
        convert = seq_len / max_possible_score
        score_counter = tabulate_score(ss)
        score_dict = self.calculate_score(score_counter, convert)
        score_dict["max_nident"] = seq_len - score_dict.get("min_mismatches")
        score_dict["nident"] = seq_len - score_dict.get("mismatches")
        score_dict["min_nident"] = seq_len - score_dict.get("max_mismatches")
        score_dict["pident"] = 100 * float(score_dict["nident"]) / seq_len
        score_dict["max_pident"] = 100 * float(score_dict["max_nident"]) / seq_len
        score_dict["min_pident"] = 100 * float(score_dict["min_nident"]) / seq_len
        score_dict["length"] = seq_len
        score_dict["evalue"] = self.evalue(score_dict["score"], seq_len)
        score_dict["pvalue"] = self.pvalue(score_dict["evalue"])
        score_dict["log_evalue"] = round(self.log_evalue(score_dict["score"], seq_len), 2)
        score_dict["log_pvalue"] = round(self.log_pvalue(score_dict["log_evalue"]), 2)
        return score_dict

    def bitscore(self, s):
        scored = self.score(s)
        score = scored.get("score")
        l = self.LAMBDA_UNGAPPED
        k = self.K_UNGAPPED
        return (l * score - np.log(k)) / np.log(2)

    def evalue(self, score, n):
        l = self.LAMBDA_UNGAPPED
        k = self.K_UNGAPPED
        m = self.DB_SIZE
        return k * m * n * np.exp(-l * score)

    def pvalue(self, evalue):
        return 1 - np.exp(-evalue)

    def log_evalue(self, score, n):
        m = self.DB_SIZE
        if m == 0:
            m = 1
        l = self.LAMBDA_UNGAPPED
        k = self.K_UNGAPPED
        return round(np.log10(k * m * n) - l * score, 2)

    def log_pvalue(self, log_evalue):
        evalue = 10 ** log_evalue
        if 1 - np.exp(-evalue) > 0:
            logp = np.log10(1 - np.exp(-evalue))
        else:
            logp = -np.inf
        if logp == -np.inf:
            return round(log_evalue, 2)
        else:
            return round(logp, 2)



import mmh3
from bitarray import bitarray


def _hash(element, seed, m):
    return mmh3.hash(element, seed) % m


def generate_hashes(element, number_hash_functions, bloomfilter_size):
    hashes = {
        _hash(element, seed, bloomfilter_size) for seed in range(number_hash_functions)
    }
    return hashes


class BloomFilter(object):
    def __init__(self, m, h):
        self.m = m
        self.h = h
        self.bitarray = bitarray(self.m)

    def __hashes(self, element):
        return generate_hashes(element, self.h, self.m)

    def add(self, e):
        for i in self.__hashes(e):
            self.bitarray[i] = True

    def update(self, elements):
        for e in list(elements):
            self.add(e)
        return self


def load_bitarray(f):
    bloomfilter = bitarray()
    with open(f, "rb") as inf:
        bloomfilter.fromfile(inf)
    return bloomfilter



DEFAULT_ROCKS_DB_STORAGE_CONFIG = {
    "filename": "test-rocksdb",
    "options": {"max_open_files": 5000, "create_if_missing": True},
}

DEFAULT_BERKELEY_DB_STORAGE_CONFIG = {"filename": "test-berkeleydb"}

DEFAULT_REDIS_STORAGE_CONFIG = {"host": "localhost", "port": 6379}
DEFAULT_PARAMETERS = {"k": 31, "m": 25 * 10 ** 6, "h": 3}
DEFAULT_ROCKS_DB_CONFIG = {
    "storage-engine": "rocksdb",
    "storage-config": DEFAULT_ROCKS_DB_STORAGE_CONFIG,
    **DEFAULT_PARAMETERS,
}

DEFAULT_REDIS_CONFIG = {
    "storage-engine": "redis",
    "storage-config": DEFAULT_REDIS_STORAGE_CONFIG,
    **DEFAULT_PARAMETERS,
}

DEFAULT_BERKELEY_DB_CONFIG = {
    "storage-engine": "berkeleydb",
    "storage-config": DEFAULT_BERKELEY_DB_STORAGE_CONFIG,
    **DEFAULT_PARAMETERS,
}

DEFAULT_CONFIG = DEFAULT_ROCKS_DB_CONFIG
DEFAULT_NPROC = 4



#! /usr/bin/env python
from __future__ import print_function
from bigsi.graph import BIGSI
import os.path
import logging
import json

logger = logging.getLogger(__name__)
from bigsi.utils import DEFAULT_LOGGING_LEVEL

logger.setLevel(DEFAULT_LOGGING_LEVEL)
from bigsi.utils import seq_to_kmers


def bloom_file_name(f):
    f = os.path.realpath(f)
    return os.path.join(f, "_".join([os.path.basename(f)]))


def bloom(config, outfile, kmers):
    outfile = os.path.realpath(outfile)
    bloomfilter = BIGSI.bloom(config, kmers)
    off = bloom_file_name(outfile)
    directory = os.path.dirname(off)
    if not os.path.exists(directory):
        os.makedirs(directory)
    with open(off, "wb") as of:
        bloomfilter.tofile(of)



from __future__ import print_function
from functools import wraps
from bigsi.utils import convert_query_kmers
from bigsi.utils import convert_query_kmer
import logging
import time
import collections
import itertools

logging.basicConfig(level=logging.DEBUG)

logger = logging.getLogger(__name__)
from bigsi.utils import DEFAULT_LOGGING_LEVEL

logger.setLevel(DEFAULT_LOGGING_LEVEL)


def choose_convert_func(kmers):
    if isinstance(kmers, str):
        convert_func = convert_query_kmer
    else:
        convert_func = convert_query_kmers
    return convert_func


def kmers_or_bytes(self, kmers):
    if self.binary_kmers and not isinstance(kmers, str):
        return [self._kmer_to_bytes(k) for k in kmers]
    elif self.binary_kmers:
        return self._kmer_to_bytes(kmers)
    else:
        return kmers


def convert_kmers(func):
    # Wrapper for functions of the form (self, kmers, colour, min_lexo=False)
    # or (self, kmers, min_lexo=False) and returns only the min of the kmer
    # and it's reverse complement
    @wraps(func)
    def convert_kmers_inner(self, kmers, *args, **kwargs):
        convert_func = choose_convert_func(kmers)
        # Are the kmers already converted
        if not kwargs.get("min_lexo"):
            # It is a list of kmers or a single kmer?
            kmers = convert_func(kmers)
        # kmers = kmers_or_bytes(self, kmers)
        return func(self, kmers, *args, **kwargs)

    return convert_kmers_inner


def convert_kmers_to_canonical(func):
    # Wrapper for functions of the form (self, kmers, colour, min_lexo=False)
    # or (self, kmers, min_lexo=False) and returns only the min of the kmer
    # and it's reverse complement
    @wraps(func)
    def convert_kmers_inner(self, kmers, *args, **kwargs):
        # logger.debug("Converting kmers to canonical")
        convert_func = choose_convert_func(kmers)
        # Are the kmers already converted
        if not kwargs.get("canonical"):
            # It is a list of kmers or a single kmer?
            kmers = convert_func(kmers)
        return func(self, kmers, *args, **kwargs)

    return convert_kmers_inner



import logging
import math
from multiprocessing import Pool
import numpy as np
from bigsi.constants import DEFAULT_CONFIG
from bigsi.graph.metadata import SampleMetadata
from bigsi.graph.index import KmerSignatureIndex
from bigsi.decorators import convert_kmers_to_canonical
from bigsi.bloom import BloomFilter
from bigsi.utils import convert_query_kmers
from bigsi.utils import seq_to_kmers
from bigsi.utils import bitwise_and
from bigsi.utils import non_zero_bitarrary_positions
from bigsi.storage import get_storage
from bigsi.constants import DEFAULT_NPROC
from collections import OrderedDict

logger = logging.getLogger(__name__)


def validate_build_params(bloomfilters, samples):
    if not len(bloomfilters) == len(samples):
        raise ValueError(
            "There must be the same number of bloomfilters and sample names"
        )


MIN_UNIQUE_KMERS_IN_QUERY = 0

B_ONE = (1).to_bytes(1, byteorder="big")


def unpack_and_sum(bitarrays):
    c = 0
    for bitarry in bitarrays:
        if c == 0:
            cumsum = np.fromstring(bitarry.unpack(one=B_ONE), dtype="i1").astype("i4")
        else:
            l = np.fromstring(bitarry.unpack(one=B_ONE), dtype="i1").astype("i4")
            cumsum = np.add(cumsum, l)
        c += 1
    return cumsum


def chunks(l, n):
    n = max(1, n)
    return (l[i : i + n] for i in range(0, len(l), n))


def unpack_bitarrays(bitarrays, j):
    logger.debug("ncores: %i" % j)
    if j == 0:
        res = unpack_and_sum(bitarrays)
        return res
    else:
        n = math.ceil(float(len(bitarrays)) / j)
        p = Pool(j)
        res = p.map(unpack_and_sum, chunks(bitarrays, n))
        p.close()
        return np.sum(res, axis=0)


import json


class BigsiQueryResult:
    PERCENT_KMERS_FOUND_KEY = "percent_kmers_found"
    NUM_KMERS_KEY = "num_kmers"
    NUM_KMERS_FOUND_KEY = "num_kmers_found"
    SAMPLE_KEY = "sample_name"

    def __init__(self, sample_name, num_kmers_found, num_kmers):
        self.sample_name = sample_name
        self.num_kmers_found = num_kmers_found
        self.num_kmers = num_kmers
        self.percent_kmers_found = round(100 * float(num_kmers_found) / num_kmers, 2)

    def todict(self):
        return {
            self.PERCENT_KMERS_FOUND_KEY: self.percent_kmers_found,
            self.NUM_KMERS_KEY: self.num_kmers,
            self.NUM_KMERS_FOUND_KEY: self.num_kmers_found,
            self.SAMPLE_KEY: self.sample_name,
        }

    def tojson(self):
        return json.dumps(self.todict())

    def __repr__(self):
        return self.tojson()

    def __eq__(self, ob):
        return self.todict() == ob.todict()


class BIGSI(SampleMetadata, KmerSignatureIndex):
    def __init__(self, config=None):
        if config is None:
            config = DEFAULT_CONFIG
        self.config = config
        self.storage = get_storage(config)
        SampleMetadata.__init__(self, self.storage)
        KmerSignatureIndex.__init__(self, self.storage)
        self.min_unique_kmers_in_query = (
            MIN_UNIQUE_KMERS_IN_QUERY
        )  ## TODO this can be inferred and set at build time

    @property
    def kmer_size(self):
        return self.config["k"]

    @property
    def nproc(self):
        return self.config.get("nproc", DEFAULT_NPROC)

    @classmethod
    def bloom(cls, config, kmers):
        kmers = convert_query_kmers(kmers)  ## Convert to canonical kmers
        bloomfilter = BloomFilter(m=config["m"], h=config["h"])
        bloomfilter.update(kmers)
        return bloomfilter.bitarray

    @classmethod
    def build(cls, config, bloomfilters, samples):
        storage = get_storage(config)
        validate_build_params(bloomfilters, samples)
        logger.debug("Insert sample metadata")
        sm = SampleMetadata(storage).add_samples(samples)
        logger.debug("Create signature index")
        ksi = KmerSignatureIndex.create(
            storage,
            bloomfilters,
            config["m"],
            config["h"],
            config.get("low_mem_build", False),
        )
        storage.close()  ## Need to delete LOCK files before re init
        return cls(config)

    def search(self, seq, threshold=1.0, score=False):
        self.__validate_search_query(seq)
        assert threshold <= 1
        kmers = set(self.seq_to_kmers(seq))
        kmers_to_colours = self.lookup(kmers)
        min_kmers = math.ceil(len(kmers) * threshold)
        if threshold == 1.0:
            results = self.exact_filter(kmers_to_colours)
        else:
            results = self.inexact_filter(kmers_to_colours, min_kmers)
        return results

    def exact_filter(self, kmers_to_colours):
        colours_with_all_kmers = non_zero_bitarrary_positions(
            bitwise_and(kmers_to_colours.values())
        )
        samples = self.get_sample_list(colours_with_all_kmers)
        return [
            BigsiQueryResult(
                sample_name=s,
                num_kmers=len(kmers_to_colours),
                num_kmers_found=len(kmers_to_colours),
            ).todict()
            for s in samples
        ]

    def get_sample_list(self, colours):
        colours_to_samples = self.colours_to_samples(colours)
        return [colours_to_samples[i] for i in colours]

    def inexact_filter(self, kmers_to_colours, min_kmers):
        num_kmers = unpack_bitarrays(list(kmers_to_colours.values()), self.nproc)
        colours = range(self.num_samples)
        colours_to_kmers_found = dict(zip(colours, num_kmers))
        colours_to_kmers_found_above_threshold = self.__colours_above_threshold(
            colours_to_kmers_found, min_kmers
        )
        results = [
            BigsiQueryResult(
                sample_name=self.colour_to_sample(colour),
                num_kmers_found=int(num_kmers_found),
                num_kmers=len(kmers_to_colours),
            )
            for colour, num_kmers_found in colours_to_kmers_found_above_threshold.items()
        ]
        results.sort(key=lambda x: x.num_kmers_found, reverse=True)
        return [r.todict() for r in results]

    def __colours_above_threshold(self, colours_to_percent_kmers, min_kmers):
        return {k: v for k, v in colours_to_percent_kmers.items() if v > min_kmers}

    def insert(self, bloomfilter, sample):
        logger.warning("Build and merge is preferable to insert in most cases")
        colour = self.add_sample(sample)
        self.insert_bloom(bloomfilter, colour - 1)

    def delete(self):
        self.storage.delete_all()

    def __validate_merge(self, bigsi):
        assert self.bloomfilter_size == bigsi.bloomfilter_size
        assert self.num_hashes == bigsi.num_hashes
        assert self.kmer_size == bigsi.kmer_size

    def merge(self, bigsi):
        self.__validate_merge(bigsi)
        self.merge_indexes(bigsi)
        self.merge_metadata(bigsi)

    # def __search_kmers(self, kmers, threshold=1, score=False):
    #     if threshold == 1:
    #         ## Special case optimisation when T==100% (we don't need to unpack the bit arrays)
    #         return self.__search_kmers_threshold_1(kmers, score=score)
    #     else:
    #         return self.__search_kmers_threshold_not_1(
    #             kmers, threshold=threshold, score=score
    #         )

    def __validate_search_query(self, seq):
        kmers = set()
        for k in self.seq_to_kmers(seq):
            kmers.add(k)
            if len(kmers) > self.min_unique_kmers_in_query:
                return True
        else:
            logger.warning(
                "Query string should contain at least %i unique kmers. Your query contained %i unique kmers, and as a result the false discovery rate may be high. In future this will become an error."
                % (self.min_unique_kmers_in_query, len(kmers))
            )

    def seq_to_kmers(self, seq):
        return seq_to_kmers(seq, self.kmer_size)

    # def __search_kmers_threshold_not_1(self, kmers, threshold, score):
    #     # if score:
    #     #     return self.__search_kmers_threshold_not_1_with_scoring(kmers, threshold)
    #     # else:
    #     return self.__search_kmers_threshold_not_1_without_scoring(kmers, threshold)

    # def __search_kmers_threshold_not_1_without_scoring(self, kmers, threshold):
    #     out = {}
    #     col_sum = self.storage.batch_lookup_and_sum_presence(kmers)
    #     for i, f in enumerate(col_sum):
    #         res = float(f) / len(kmers)
    #         if res >= threshold:
    #             sample = self.storage.colour_to_sample(i)
    #             if sample != "DELETED":
    #                 out[sample] = {}
    #                 out[sample]["percent_kmers_found"] = 100 * res
    #     return out

    # def __search_kmers_threshold_1(self, kmers, score=False):
    #     """Special case where the threshold is 1 (can accelerate queries with AND)"""
    #     bitarray = self.storage.batch_lookup_and_bitwise_and_presence(kmers)
    #     out = {}
    #     for c in bitarray:
    #         sample = self.storage.colour_to_sample(c)
    #         if sample != "DELETED":
    #             if score:
    #                 out[sample] = self.scorer.score(
    #                     "1" * (len(kmers) + self.kmer_size - 1)
    #                 )  # Fix!
    #             else:
    #                 out[sample] = {}
    #             out[sample]["percent_kmers_found"] = 100
    #     return out



from bigsi.storage.base import BaseStorage
from bigsi.constants import DEFAULT_REDIS_STORAGE_CONFIG
import redis
from bitarray import bitarray
from bigsi.utils import batch


class RedisStorage(BaseStorage):
    def __init__(self, storage_config=None):
        if storage_config is None:
            storage_config = DEFAULT_REDIS_STORAGE_CONFIG
        self.storage_config = storage_config
        self.storage = redis.StrictRedis(**storage_config)
        self.pipe = self.storage.pipeline()
        self.write_batch_size = int(self.storage_config.get("write_batch_size", 10000))

    def __repr__(self):
        return "redis Storage"

    def delete_all(self):
        return self.storage.flushall()

    def __execute_pipeline(self):
        result = self.pipe.execute()
        self.pipe = self.storage.pipeline()
        return result

    def batch_set(self, keys, values):
        for batchiter in batch(zip(keys, values), self.write_batch_size):
            for k, v in batchiter:
                self.pipe.set(k, v)
            self.__execute_pipeline()

    def batch_get(self, keys):
        for k in keys:
            self.pipe.get(k)
        return self.__execute_pipeline()

    def set_bit(self, key, pos, bit):
        _key = self.convert_to_bitarray_key(key)
        self.storage.setbit(_key, pos, bit)

    def get_bit(self, key, pos):
        _key = self.convert_to_bitarray_key(key)
        return bool(self.storage.getbit(_key, pos))

    def incr(self, key):
        __key = self.convert_to_integer_key(key)
        return self.storage.incr(__key)



from bigsi.storage.berkeleydb import BerkeleyDBStorage
from bigsi.storage.redis import RedisStorage
from bigsi.storage.rocksdb import RocksDBStorage


def get_storage(config):
    return {
        "rocksdb": RocksDBStorage,
        "berkeleydb": BerkeleyDBStorage,
        "redis": RedisStorage,
    }[config["storage-engine"]](config["storage-config"])



import logging

from bigsi.bloom import generate_hashes
from bigsi.bloom import BloomFilter
from bigsi.matrix import transpose
from bigsi.matrix import BitMatrix
from bigsi.utils import convert_query_kmer
from bigsi.utils import bitwise_and

BLOOMFILTER_SIZE_KEY = "ksi:bloomfilter_size"
NUM_HASH_FUNCTS_KEY = "ksi:num_hashes"
logger = logging.getLogger(__name__)


class KmerSignatureIndex:

    """
    Methods for managing kmer signature indexes
    """

    def __init__(self, storage):
        self.storage = storage
        self.bitmatrix = BitMatrix(storage)
        self.bloomfilter_size = storage.get_integer(BLOOMFILTER_SIZE_KEY)
        self.num_hashes = storage.get_integer(NUM_HASH_FUNCTS_KEY)

    @classmethod
    def create(cls, storage, bloomfilters, bloomfilter_size, num_hashes, lowmem=False):
        bloomfilters = [
            bf.bitarray if isinstance(bf, BloomFilter) else bf for bf in bloomfilters
        ]
        storage.set_integer(BLOOMFILTER_SIZE_KEY, bloomfilter_size)
        storage.set_integer(NUM_HASH_FUNCTS_KEY, num_hashes)
        logger.debug("Transpose bitarrays")
        rows = transpose(bloomfilters, lowmem=lowmem)
        logger.debug("Insert rows")
        bitmatrix = BitMatrix.create(
            storage, rows, num_rows=bloomfilter_size, num_cols=len(bloomfilters)
        )
        return cls(storage)

    def lookup(self, kmers):
        if isinstance(kmers, str):
            kmers = [kmers]
        kmer_to_hashes = self.__kmers_to_hashes(kmers)
        hashes = {h for sublist in kmer_to_hashes.values() for h in sublist}
        rows = self.__batch_get_rows(hashes)
        return self.__bitwise_and_kmers(kmer_to_hashes, rows)

    def insert_bloom(self, bloomfilter, column_index):
        self.bitmatrix.insert_column(bloomfilter, column_index)

    def merge_indexes(self, ksi):
        for i in range(self.bloomfilter_size):
            r1 = self.bitmatrix.get_row(i)
            r2 = ksi.bitmatrix.get_row(i)
            r1.extend(r2)
            self.bitmatrix.set_row(i, r1)
        self.bitmatrix.set_num_cols(self.bitmatrix.num_cols + ksi.bitmatrix.num_cols)

    def __kmers_to_hashes(self, kmers):
        d = {}
        for k in set(kmers):
            d[k] = set(
                generate_hashes(
                    convert_query_kmer(k), self.num_hashes, self.bloomfilter_size
                )
            )  ## use canonical kmer to generate lookup, but report query kmer
        return d

    def __batch_get_rows(self, row_indexes):
        return dict(zip(row_indexes, self.bitmatrix.get_rows(row_indexes)))

    def __bitwise_and_kmers(self, kmer_to_hashes, rows):
        d = {}
        for k, hashes in kmer_to_hashes.items():
            subset_rows = [rows[h] for h in hashes]
            d[k] = bitwise_and(subset_rows)
        return d



"""
Cortex file reader.
## Adapted with thanks from code writen by https://github.com/jeromekelleher/
"""
from __future__ import print_function
from __future__ import division

__version__ = "0.0.1"


import os
import json
import gzip
import struct
import subprocess
import math
from bigsi.utils import seq_to_kmers

BITS = {"A": "00", "G": "01", "C": "10", "T": "11"}
BASES = {"00": "A", "01": "G", "10": "C", "11": "T"}


def extract_kmers_from_ctx(ctx, k):
    gr = GraphReader(ctx)
    for i in gr:
        for kmer in seq_to_kmers(i.kmer.canonical_value, k):
            yield kmer


def kmer_to_bits(kmer):
    return "".join([BITS[k] for k in kmer])


def decode_kmer(binary_kmer, kmer_size):
    """
    Returns a string representation of the specified kmer.
    """
    # G and C are the wrong way around because we reverse the sequence.
    # This really is a nasty way to do this!
    assert kmer_size <= 31
    binary_kmer_int = struct.unpack("Q", binary_kmer)[0]

    b = "{0:064b}".format(binary_kmer_int)[::-1]
    ret = []
    for j in range(kmer_size):
        nuc = BASES[b[j * 2 : (j + 1) * 2]]
        ret.append(nuc)
    ret = "".join(ret)

    return ret[::-1]


def decode_edges(edges):
    """
    Decodes the specified integer representing edges in Cortex graph. Returns
    a tuple (forward, reverse) which contain the list of nucleotides that we
    append to a kmer and its reverse complement to obtain other kmers in the
    Cortex graph.
    """
    bases = ["A", "C", "G", "T"]
    fwd = []
    for j in range(4):
        if (1 << j) & edges != 0:
            fwd.append(bases[j])
    rev = []
    bases.reverse()
    for j in range(4):
        if (1 << (j + 4)) & edges != 0:
            rev.append(bases[j])
    return fwd, rev


def encode_kmer(kmer):
    """
    Returns the encoded integer representation of the specified string kmer.
    """
    ret = 0
    codes = {"A": 0, "C": 1, "G": 2, "T": 3}
    for j, nuc in enumerate(reversed(kmer)):
        v = codes[nuc]
        ret |= v << (2 * j)
    return struct.pack("Q", ret)


def reverse_complement(kmer):
    """
    Returns the reverse complement of the specified string kmer.
    """
    d = {"A": "T", "C": "G", "G": "C", "T": "A"}
    # This is very slow and nasty!
    s = ""
    for c in kmer:
        s += d[c]
    return s[::-1]


def canonical_kmer(kmer):
    """
    Returns the canonical version of this kmer, which is the lexically
    least of itself and its reverse complement.
    """
    rev = reverse_complement(kmer)
    return rev if rev < kmer else kmer


class Kmer(object):

    """
    A class representing fixed length strings of DNA nucleotides. A Kmer
    is equal to its reverse complement, and the canonical representation
    of a give kmer is the lexically least of itself  and its reverse
    complement.
    """

    def __init__(self, kmer):
        self.value = kmer
        self.canonical_value = canonical_kmer(kmer)

    def __str__(self):
        return self.canonical_value


class CortexRecord(object):

    """
    Class representing a single record in a cortex graph. A record
    consists of a kmer, its edges and coverages in the colours.
    """

    def __init__(
        self, kmer_size, kmer, coverages, edges, num_colours=1, binary_kmer=False
    ):
        if binary_kmer:
            self.kmer = kmer
        else:
            self.kmer = Kmer(decode_kmer(kmer, kmer_size))
        self.coverages = coverages
        self.edges = [decode_edges(e) for e in edges]
        self.num_colours = num_colours

    def __str__(self):
        return "<CortexRecord %s - %i colour(s)>" % (self.kmer, self.num_colours)

    def print(self, colour):
        nucleotides = "ACGT"
        s = ["." for j in range(8)]
        for j, n in enumerate(nucleotides):
            if n in self.edges[colour][1]:
                s[j] = n.lower()
        for j, n in enumerate(nucleotides):
            if n in self.edges[colour][0]:
                s[j + 4] = n
        s = "".join(s)
        return "{0} {1} {2}".format(self.kmer, self.coverages[colour], s)

    def get_adjacent_kmers(self, colour=0, direction=0):
        """
        Returns the kmers adjacent to this kmer using the edges in the
        record.
        """
        fwd, rev = self.edges[colour]
        if direction == 0:
            for n in fwd:
                yield Kmer(self.kmer.canonical_value[1:] + n)
        else:
            for n in rev:
                yield Kmer(n + self.kmer.canonical_value[:-1])


class GraphReader(object):

    """
    Class to read a cortex graph.
    """

    def __init__(self, graph_file, binary_kmers=False):
        self._file_name = graph_file
        self._file = open(graph_file, "rb")
        self.read_header()
        self.binary_kmers = binary_kmers  # report kmers as bytes

    def read_unsigned_int(self):
        """
        Reads an uint32_t from the stream.
        """
        b = self._file.read(4)
        return struct.unpack("<I", b)[0]

    def read_header(self):
        """
        Reads the header of the graph file.
        """
        magic_str = b"CORTEX"
        b = self._file.read(len(magic_str))
        if b != magic_str:
            raise ValueError("File format mismatch")
        self.version = self.read_unsigned_int()
        if self.version != 6:
            raise ValueError("File format version error; only 6 supported")
        self.kmer_size = self.read_unsigned_int()
        self.kmer_storage_size = 8 * self.read_unsigned_int()
        self.num_colours = self.read_unsigned_int()
        self.record_size = self.kmer_storage_size + 5 * self.num_colours
        # skip per colour mean_read_length and total_sequence
        skip = self.num_colours * 12
        self._file.seek(skip, os.SEEK_CUR)
        for j in range(self.num_colours):
            v = self.read_unsigned_int()
            self._file.seek(v, os.SEEK_CUR)
        # skip per colour error rates
        skip = self.num_colours * 16  # sizeof(long double)
        self._file.seek(skip, os.SEEK_CUR)
        for j in range(self.num_colours):
            # skip cleaning counters
            self._file.seek(12, os.SEEK_CUR)
            v = self.read_unsigned_int()
            self._file.seek(v, os.SEEK_CUR)
        # header ends with the magic word
        b = self._file.read(len(magic_str))
        if b != magic_str:
            raise ValueError("File format mismatch")
        payload_start = self._file.tell()
        self._file.seek(0, os.SEEK_END)
        payload_size = self._file.tell() - payload_start
        self.num_records = payload_size // self.record_size
        self._file.seek(payload_start, os.SEEK_SET)

    def __iter__(self):
        return self

    def __next__(self):
        """
        Returns the next record
        """
        buf = self._file.read(self.record_size)
        if len(buf) == 0:
            raise StopIteration()
        return self.decode_record(buf)

    def next(self):
        # Python 2 compat - TODO what is the recommended way to do this?
        return self.__next__()

    def decode_record(self, buff):
        """
        Decodes the specified graph record.
        """

        kmer = buff[0:8]  #
        # print(buff[0:8], kmer)
        offset = 8
        coverages = struct.unpack_from("I" * self.num_colours, buff, offset)
        offset += self.num_colours * 4
        edges = struct.unpack_from("B" * self.num_colours, buff, offset)
        # print(edges)
        record = CortexRecord(
            self.kmer_size,
            kmer,
            coverages,
            edges,
            num_colours=self.num_colours,
            binary_kmer=self.binary_kmers,
        )
        return record


class LinksRecord(object):

    """
    Class representing a single links record attached to a kmer.
    """

    def __init__(self, direction, num_kmers, counts, junctions):
        self.direction = direction
        self.num_kmers = num_kmers
        self.counts = counts
        self.junctions = junctions

    def __str__(self):
        return "{0}:{1}:{2}:{3}".format(
            self.direction, self.num_kmers, self.counts, self.junctions
        )


class LinksFile(object):

    """
    Class to read a cortex links file.
    """

    def __init__(self, filename):
        self._file = gzip.open(filename, "r")
        self._read_header()

    def _read_header(self):
        """
        Reads the header of the links file and a parses the relevant fields.
        """
        not_done = True
        open_braces = 0
        closed_braces = 0
        header = ""
        while not_done:
            s = self._file.readline()
            open_braces += s.count("{")
            closed_braces += s.count("}")
            not_done = open_braces != closed_braces
            header += s
        metadata = json.loads(header)
        if "fileFormat" in metadata:
            assert metadata["fileFormat"] == "ctp"
            assert metadata["formatVersion"] == 2
            self.num_kmers_with_paths = metadata["num_kmers_with_paths"]
            self.num_paths = metadata["num_paths"]
            self.ncols = metadata["ncols"]
            self.kmer_size = metadata["kmer_size"]
            self.num_kmers_in_graph = metadata["num_kmers_in_graph"]
            self.colours = metadata["colours"]
            self.commands = metadata["commands"]
        else:
            assert metadata["file_format"] == "ctp"
            assert metadata["format_version"] == 3
            graph = metadata["graph"]
            self.kmer_size = graph["kmer_size"]
            self.num_colours = graph["num_colours"]
            self.num_kmers_in_graph = graph["num_kmers_in_graph"]
            self.colours = graph["colours"]
            paths = metadata["paths"]
            self.num_kmers_with_paths = paths["num_kmers_with_paths"]

    def __iter__(self):
        return self

    def __next__(self):
        """
        Return the next record.
        """
        # Skip empty lines and comments
        s = self._file.readline()
        if s == "":
            raise StopIteration()
        s = s.lstrip()
        while len(s) == 0 or s.startswith("#"):
            s = self._file.readline()
            if s == "":
                raise StopIteration()
            s = s.lstrip()
        # Now read the header for the kmer.
        split = s.split()
        kmer = split[0]
        num_paths = int(split[1])
        paths = []
        for j in range(num_paths):
            split = self._file.readline().split()
            direction = split[0]
            num_kmers = int(split[1])
            num_juncs = int(split[2])
            counts = [int(x) for x in split[3].split(",")]
            juncs = split[4]
            assert num_juncs == len(juncs)
            lr = LinksRecord(direction, num_kmers, counts, juncs)
            paths.append(lr)
        return kmer, paths

    def next(self):
        """
        Python 2.x compatability.
        TODO check the recommended way to do this in 3.x and 2.x.
        """
        return self.__next__()


class GraphTraverser(object):

    """
    Class to traverse a graph using links information.
    """

    def __init__(self, graph_reader, links_file):
        self._graph = {}
        for r in graph_reader:
            self._graph[r.kmer.canonical_value] = r
        self._links = {}
        for kmer, link_records in links_file:
            assert len(kmer) == links_file.kmer_size
            self._links[kmer] = link_records
        assert links_file.num_kmers_with_paths == len(self._links)

    def traverse(self, seed):
        """
        Performs a simple traversal for the specified seed, returning the
        resulting contig.
        """
        # print("seed = ", seed)
        k = seed
        contig = seed[:-1]
        o = 0
        paths = []
        while k is not None:
            b = k[-1] if o == 0 else reverse_complement(k[0])
            contig += b
            revcmp = reverse_complement(k)
            c = k
            if revcmp < k:
                o = (o + 1) % 2
                c = revcmp
            direction = "F" if o == 0 else "R"
            if c in self._links:
                for lr in self._links[c]:
                    if lr.direction == direction:
                        paths.append([0, lr.junctions])
            adj = [obj.value for obj in self._graph[c].get_adjacent_kmers(0, o)]
            # print(k, c, o, adj, contig)
            # print("paths = ")
            # for p in paths:
            #     print("\t", p)
            k = None
            if len(adj) == 1:
                k = adj[0]
            elif len(adj) > 1:
                junctions = paths[0][1]
                # Choose the branch, if we can.
                junction = junctions[0]
                j = -1
                b = junction
                if o != 0:
                    j = 0
                    b = reverse_complement(junction)
                for kp in adj:
                    if kp[j] == b:
                        k = kp
                if k is None:
                    print("no junctions matched")
                # update the paths
                oldpaths = paths
                paths = []
                for age, junctions in oldpaths:
                    if junctions[0] == junction and len(junctions) > 1:
                        # print("updating", age, junctions)
                        paths.append([age + 1, junctions[1:]])

        return contig


def run_cortex(command, args, path="./ctx31"):
    """
    Runs the specified command on cortex, raising an error if the exit status
    is non zero.
    """
    cmd = [path, command, "-q", "-f", "-m", "100M", "-t", "10"] + args
    subprocess.check_call(cmd)


def build_graph(kmer_size, fasta_file, cortex_file, links_file=None):
    """
    Builds a graph from the specified fasta file and writes it to the
    specified cortex file. If the links_file is not None, also run
    thread to thread the sequences through the graph.
    """
    cmd = ["-s", "sample", "-k", str(kmer_size), "-1", fasta_file, cortex_file]
    run_cortex("build", cmd)
    if links_file is not None:
        cmd = ["-1", fasta_file, "-o", links_file, cortex_file]
        run_cortex("thread", cmd)



from bitarray import bitarray

NUM_ROWS_KEY = "number_of_rows"
NUM_COLS_KEY = "number_of_cols"


class BitMatrix(object):

    """
    Manages the gets and sets of the bitmatrix to the various storage backends. 
    Does not know the concept of a kmer.
    """

    def __init__(self, storage):
        self.storage = storage
        self.num_rows = self.storage.get_integer(NUM_ROWS_KEY)
        self.num_cols = self.storage.get_integer(NUM_COLS_KEY)

    @classmethod
    def create(cls, storage, rows, num_rows, num_cols):
        storage.set_bitarrays(range(num_rows), rows)
        storage.set_integer(NUM_ROWS_KEY, num_rows)
        storage.set_integer(NUM_COLS_KEY, num_cols)
        storage.sync()
        return cls(storage)

    def get_row(self, row_index):
        return self.storage.get_bitarray(row_index)[: self.num_cols]

    def get_rows(self, row_indexes):
        # Takes advantage of batching in storage engine if available
        return (ba[: self.num_cols] for ba in self.storage.get_bitarrays(row_indexes))

    def set_row(self, row_index, bitarray):
        return self.storage.set_bitarray(row_index, bitarray)

    def set_rows(self, row_indexes, bitarrays):
        # Takes advantage of batching in storage engine if available
        return self.storage.set_bitarrays(row_indexes, bitarrays)

    def set_num_cols(self, num_cols):
        self.num_cols = num_cols
        self.storage.set_integer(NUM_COLS_KEY, self.num_cols)

    def get_column(self, column_index):
        ## This is very slow, as we index row-wise. Need to know the number of rows, so must be done elsewhere
        return bitarray(
            "".join(
                [
                    str(int(i))
                    for i in self.storage.get_bits(
                        list(range(self.num_rows)), [column_index] * self.num_rows
                    )
                ]
            )
        )

    def get_columns(self, column_indexes):
        for column_index in column_indexes:
            yield self.get_column(column_index)

    def insert_column(self, bitarray, column_index):
        ## This is very slow, as we index row-wise
        self.storage.set_bits(
            list(range(len(bitarray))),
            [column_index] * len(bitarray),
            bitarray.tolist(),
        )
        if column_index >= self.num_cols:
            self.set_num_cols(self.num_cols + 1)



#! /usr/bin/env python
from __future__ import print_function
from bigsi.graph import BIGSI
import os.path
import sys
import logging
import json

logger = logging.getLogger(__name__)
from bigsi.utils import DEFAULT_LOGGING_LEVEL

logger.setLevel(DEFAULT_LOGGING_LEVEL)

import numpy as np
from bitarray import bitarray
import math
from bigsi.utils import chunks
from bigsi.cmds.bloom import bloom_file_name
import tempfile


def load_bloomfilter(f):
    ff = bloom_file_name(f)
    logger.debug("Loading %s " % ff)
    bloomfilter = bitarray()
    with open(ff, "rb") as inf:
        bloomfilter.fromfile(inf)
    return bloomfilter


def get_required_bytes_per_bloomfilter(m):
    return m * 9 / 8


def get_required_chunk_size(N, m, max_memory):
    bytes_per_bloomfilter = get_required_bytes_per_bloomfilter(m)
    required_bytes = bytes_per_bloomfilter * N
    num_chunks = math.ceil(required_bytes / max_memory)
    chunk_size = math.floor(N / num_chunks)
    return chunk_size, num_chunks


def build(config, bloomfilter_filepaths, samples, max_memory=None):
    # Max memory is in bytes
    if max_memory is None:
        chunk_size = len(bloomfilter_filepaths)
        num_chunks = 1
    else:
        chunk_size, num_chunks = get_required_chunk_size(
            N=len(samples), m=config["h"], max_memory=max_memory
        )
    if chunk_size < 1:
        raise ValueError("Max memory must be at least 8 * Bloomfilter size in bytes")
    LL = list(zip(bloomfilter_filepaths, samples))
    for i, v in enumerate(chunks(LL, chunk_size)):
        bloomfilter_filepaths = [x[0] for x in v]
        # print(bloomfilter_filepaths)
        samples = [x[1] for x in v]
        logger.info("Building index: %i/%i" % (i, num_chunks))
        if i == 0:
            index = build_main(config, bloomfilter_filepaths, samples)
        else:
            tmp_index = build_tmp(config, bloomfilter_filepaths, samples, i)
            index.merge(tmp_index)
            tmp_index.delete()
    return {"result": "success"}


def build_main(config, bloomfilter_filepaths, samples):
    bloomfilters = []
    for f in bloomfilter_filepaths:
        bloomfilters.append(load_bloomfilter(f))
    return BIGSI.build(config, bloomfilters, samples)


import copy


def build_tmp(config, bloomfilter_filepaths, samples, i):
    tmpconfig = copy.copy(config)
    if config["storage-engine"] == "redis":
        config["storage-config"]["db"] = i
    else:
        config["storage-config"]["filename"] = config["filename"] + "%i.tmp" % i
    return build_main(config, bloomfilter_filepaths, samples, index)



import hashlib
import struct
import sys
import logging
from functools import reduce
import numpy as np
from itertools import islice, chain

logger = logging.getLogger(__name__)
logger.setLevel("DEBUG")

COMPLEMENT = {"A": "T", "C": "G", "G": "C", "T": "A"}
BITS = {"A": "00", "G": "01", "C": "10", "T": "11"}
BASES = {"00": "A", "01": "G", "10": "C", "11": "T"}


def batch(iterable, size):
    sourceiter = iter(iterable)
    while True:
        batchiter = islice(sourceiter, size)
        yield chain([next(batchiter)], batchiter)


def bitwise_and(bitarrays):
    return reduce(lambda x, y: x & y, bitarrays)


def non_zero_bitarrary_positions(bitarray):
    return np.where(bitarray)[0].tolist()


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i : i + n]


def reverse_comp(s):
    return "".join([COMPLEMENT.get(base, base) for base in reversed(s)])


def convert_query_kmers(kmers):
    for k in kmers:
        yield convert_query_kmer(k)


def convert_query_kmer(kmer):
    return canonical(kmer)


def canonical(k):
    l = [k, reverse_comp(k)]
    l.sort()
    return l[0]


def min_lexo(k):
    l = [k, reverse_comp(k)]
    l.sort()
    return l[0]


def seq_to_kmers(seq, kmer_size):
    for i in range(len(seq) - kmer_size + 1):
        yield seq[i : i + kmer_size]



from bitarray import bitarray
import struct
import gc
import logging

logger = logging.getLogger(__name__)


class BaseStorage(object):
    def convert_key_to_bytes(self, key):
        return key.encode("utf-8")

    def __setitem__(self, key, val):
        if not isinstance(key, bytes):
            key = self.convert_key_to_bytes(key)
        self.storage[key] = val

    def __getitem__(self, key):
        if not isinstance(key, bytes):
            key = self.convert_key_to_bytes(key)
        return self.storage[key]

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def convert_to_integer_key(self, key):
        return str(key) + ":int"

    def convert_to_string_key(self, key):
        return str(key) + ":string"

    def convert_to_bitarray_key(self, key):
        return str(key) + ":bitarray"

    def convert_integer_batch_keys(self, keys):
        return (
            self.convert_key_to_bytes(self.convert_to_integer_key(key)) for key in keys
        )

    def convert_bitarray_batch_keys(self, keys):
        return (
            self.convert_key_to_bytes(self.convert_to_bitarray_key(key)) for key in keys
        )

    def int_to_bytes(self, value):
        return str(value).encode("utf-8")

    def bytes_to_int(self, value):
        return int(value.decode("utf-8"))

    def batch_set(self, keys, values):
        for k, v in zip(keys, values):
            self[k] = v

    def batch_get(self, keys):
        return [self[k] for k in keys]

    def set_integer(self, key, value):
        key = self.convert_to_integer_key(key)
        self[key] = self.int_to_bytes(value)

    def get_integer(self, key):
        key = self.convert_to_integer_key(key)
        return self.bytes_to_int(self[key])

    def set_integers(self, keys, values):
        _keys = self.convert_integer_batch_keys(keys)
        self.batch_set(_keys, (self.int_to_bytes(v) for v in values))

    def get_integers(self, keys):
        _keys = self.convert_integer_batch_keys(keys)
        return [self.bytes_to_int(b) for b in self.batch_get(_keys)]

    def set_string(self, key, value):
        assert isinstance(value, str)
        key = self.convert_to_string_key(key)
        self[key] = value.encode("utf-8")

    def get_string(self, key):
        key = self.convert_to_string_key(key)
        return self[key].decode("utf-8")

    def set_bitarray(self, key, value):
        assert isinstance(value, bitarray)
        _key = self.convert_to_bitarray_key(key)
        self[_key] = value.tobytes()

    def set_bitarrays(self, keys, values):
        logger.debug("set bitarrays")
        _keys = self.convert_bitarray_batch_keys(keys)
        self.batch_set(_keys, (v.tobytes() for v in values))

    def load_bitarray(self, _bytes):
        ba = bitarray()
        ba.frombytes(_bytes)
        return ba

    def get_bitarray(self, key):
        _key = self.convert_to_bitarray_key(key)
        value = self.load_bitarray(self[_key])
        return value

    def get_bitarrays(self, keys):
        # Takes advantage of batching in storage engine if available
        _keys = self.convert_bitarray_batch_keys(keys)
        return (self.load_bitarray(result) for result in self.batch_get(_keys))

    def set_bit(self, key, pos, bit):
        ba = self.get_bitarray(key)
        try:
            ba[pos] = bit
        except IndexError:
            ba.append(bit)  ## Assuming setbit is always in the next available column
        self.set_bitarray(key, ba)

    def set_bits(self, keys, positions, bits):
        # Takes advantage of batching in storage engine if available
        for key, pos, bit in zip(keys, positions, bits):
            self.set_bit(key, pos, bit)

    def get_bit(self, key, pos):
        return self.get_bitarray(key)[pos]

    def get_bits(self, keys, positions):
        # Takes advantage of batching in storage engine if available
        for key, pos in zip(keys, positions):
            yield self.get_bit(key, pos)

    def delete_all(self):
        raise NotImplementedError("Implemented in subclass")

    def incr(self, key):
        try:
            i = self.get_integer(key)
            i += 1
            self.set_integer(key, i)
            return i
        except KeyError:
            i = 1
            self.set_integer(key, i)
            return i

    def sync(self):
        pass

    def close(self):
        del self.storage
        gc.collect()



import hypothesis.strategies as st
from bigsi import BIGSI
import os
import itertools


ROCKS_DB_STORAGE_CONFIG = {
    "filename": "test-rocksdb",
    "options": {"max_open_files": 5000, "create_if_missing": True},
}

BERKELEY_DB_STORAGE_CONFIG = {"filename": "test-berkeleydb"}

REDIS_STORAGE_CONFIG = {"host": "localhost", "port": 6379}
PARAMETERS = {"k": 3, "m": 1000, "h": 3}
ROCKS_DB_CONFIG = {
    "storage-engine": "rocksdb",
    "storage-config": ROCKS_DB_STORAGE_CONFIG,
    **PARAMETERS,
}

REDIS_CONFIG = {
    "storage-engine": "redis",
    "storage-config": REDIS_STORAGE_CONFIG,
    **PARAMETERS,
}

BERKELEY_DB_CONFIG = {
    "storage-engine": "berkeleydb",
    "storage-config": BERKELEY_DB_STORAGE_CONFIG,
    **PARAMETERS,
}


CONFIGS = [REDIS_CONFIG, ROCKS_DB_CONFIG, BERKELEY_DB_CONFIG]
L = ["".join(x) for x in itertools.product("ATCG", repeat=9)]
ST_KMER = st.sampled_from(L)
# ST_KMER_SIZE = st.integers(min_value=11, max_value=31)
ST_KMER_SIZE = st.just(13)

ST_SEQ = st.text(min_size=31, max_size=1000, alphabet=["A", "T", "C", "G"])

ST_SAMPLE_NAME = st.text(min_size=1)
# ST_GRAPH = st.just(BIGSI)
ST_SAMPLE_COLOUR = st.integers(min_value=0, max_value=10)
ST_BLOOM_FILTER_SIZE = st.integers(min_value=10, max_value=100)
ST_NUM_HASHES = st.integers(min_value=10, max_value=100)


BINARY_KMERS_OR_NOT = [True, False]
ST_BINARY_KMERS = st.just(False)  # st.sampled_from(BINARY_KMERS_OR_NOT)



DELETION_SPECIAL_SAMPLE_NAME = "D3L3T3D"


class SampleMetadata:
    def __init__(self, storage):
        self.storage = storage

    @property
    def colour_count_key(self):
        return "colour_count"

    @property
    def num_samples(self):
        try:
            return self.__get_integer(self.colour_count_key)
            # we could distinguish between the number of samples,
            # and the number of colours,
            # but it adds unuseful complexity
        except KeyError:
            return 0

    def add_sample(self, sample_name):
        self.__validate_sample_name(sample_name)
        colour = self.num_samples
        self.__set_sample_colour(sample_name, colour)
        self.__set_colour_sample(colour, sample_name)
        return self.__increment_colour_count()

    def add_samples(self, sample_names):
        for sample_name in sample_names:
            self.add_sample(sample_name)

    def delete_sample(self, sample_name):
        ## Deleting samples just changes it's name to a reserved deleted string
        colour = self.sample_to_colour(sample_name)
        self.__set_colour_sample(colour, DELETION_SPECIAL_SAMPLE_NAME)
        self.__set_sample_colour(sample_name, -1)
        ## We don't decrement the count, as the number of colours is the same

    def sample_name_exists(self, sample_name):
        try:
            self.__get_integer(sample_name)
            return True
        except KeyError:
            return False

    def sample_to_colour(self, sample_name):
        try:
            colour = self.__get_integer(sample_name)
            if colour < 0:
                return None
            else:
                return colour
        except KeyError:
            return None

    def colour_to_sample(self, colour):
        ## Ignores deleted samples
        sample_name = self.__get_string(colour)
        if sample_name == DELETION_SPECIAL_SAMPLE_NAME:
            sample_name = "DELETED_SAMPLE"
        return sample_name

    def samples_to_colours(self, sample_names):
        return {
            s: self.sample_to_colour(s)
            for s in sample_names
            if self.sample_to_colour(s) is not None
        }

    def colours_to_samples(self, colours):
        return {
            c: self.colour_to_sample(c) for c in colours if self.colour_to_sample(c)
        }

    def merge_metadata(self, sm):
        for c in range(sm.num_samples):
            sample = sm.colour_to_sample(c)
            try:
                self.add_sample(sample)
            except ValueError:
                self.add_sample(sample + "_duplicate_in_merge")

    def __set_integer(self, key, value):
        __key = self.__add_key_prefix(key)
        self.storage.set_integer(__key, value)

    def __get_integer(self, key):
        __key = self.__add_key_prefix(key)
        return self.storage.get_integer(__key)

    def __set_string(self, key, value):
        __key = self.__add_key_prefix(key)
        self.storage.set_string(__key, value)

    def __get_string(self, key):
        __key = self.__add_key_prefix(key)
        return self.storage.get_string(__key)

    def __incr(self, key):
        __key = self.__add_key_prefix(key)
        return self.storage.incr(__key)

    def __set_sample_colour(self, sample_name, colour):
        self.__set_integer(sample_name, colour)

    def __set_colour_sample(self, colour, sample_name):
        self.__set_string(colour, sample_name)

    def __increment_colour_count(self):
        return self.__incr(self.colour_count_key)

    def __add_key_prefix(self, key):
        return ":".join(["metadata", str(key)])

    def __validate_sample_name(self, sample_name):
        if sample_name == DELETION_SPECIAL_SAMPLE_NAME:
            raise ValueError(
                "You can't call a sample %s" % DELETION_SPECIAL_SAMPLE_NAME
            )
        if self.sample_name_exists(sample_name):
            raise ValueError("You can't insert two samples with the same name")



#! /usr/bin/env python
from __future__ import print_function
import sys
import os
import argparse
import redis
import json
import math
import logging
import hug
import tempfile
import humanfriendly
import yaml

from bigsi.version import __version__
from bigsi.graph import BIGSI

from bigsi.cmds.insert import insert
from bigsi.cmds.search import search
from bigsi.cmds.delete import delete
from bigsi.cmds.bloom import bloom
from bigsi.cmds.build import build
from bigsi.cmds.merge import merge

from bigsi.storage import get_storage

from bigsi.utils.cortex import extract_kmers_from_ctx
from bigsi.utils import seq_to_kmers


logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


API = hug.API("bigsi-%s" % str(__version__))


def get_config_from_file(config_file):
    if config_file is None:
        if os.environ.get("BIGSI_CONFIG"):
            config_file = os.environ.get("BIGSI_CONFIG")
        else:
            return DEFAULT_CONFIG
    with open(config_file, "r") as infile:
        config = yaml.load(infile)
    return config


@hug.object(name="bigsi", version="0.1.1", api=API)
@hug.object.urls("/", requires=())
class bigsi(object):
    @hug.object.cli
    @hug.object.post("/insert", output_format=hug.output_format.json)
    def insert(self, config: hug.types.text, bloomfilter, sample):
        """Inserts a bloom filter into the graph

        e.g. bigsi insert ERR1010211.bloom ERR1010211

        """
        config = get_config_from_file(config)
        index = BIGSI(config)
        return insert(index=index, bloomfilter=bloomfilter, sample=sample)

    @hug.object.cli
    @hug.object.post("/bloom")
    def bloom(self, ctx, outfile, config=None):
        """Creates a bloom filter from a sequence file or cortex graph. (fastq,fasta,bam,ctx)

        e.g. index insert ERR1010211.ctx

        """
        config = get_config_from_file(config)
        bf = bloom(
            config=config,
            outfile=outfile,
            kmers=extract_kmers_from_ctx(ctx, config["k"]),
        )

    @hug.object.cli
    @hug.object.post("/build", output_format=hug.output_format.json)
    def build(
        self,
        bloomfilters: hug.types.multiple,
        samples: hug.types.multiple = [],
        config: hug.types.text = None,
    ):
        config = get_config_from_file(config)

        if samples:
            assert len(samples) == len(bloomfilters)
        else:
            samples = bloomfilters

        if config.get("max_build_mem_bytes"):
            max_memory_bytes = humanfriendly.parse_size(config["max_build_mem_bytes"])
        else:
            max_memory_bytes = None

        return build(
            config=config,
            bloomfilter_filepaths=bloomfilters,
            samples=samples,
            max_memory=max_memory_bytes,
        )

    @hug.object.cli
    @hug.object.post("/merge", output_format=hug.output_format.json)
    def merge(self, config: hug.types.text, merge_config: hug.types.text):
        config = get_config_from_file(config)
        merge_config = get_config_from_file(merge_config)
        index1 = BIGSI(config)
        index2 = BIGSI(merge_config)
        merge(index1, index2)
        return {"result": "merged %s into %s." % (merge_config, config)}

    @hug.object.cli
    @hug.object.get(
        "/search",
        examples="seq=ACACAAACCATGGCCGGACGCAGCTTTCTGA",
        output_format=hug.output_format.json,
        response_headers={"Access-Control-Allow-Origin": "*"},
    )
    def search(
        self,
        seq: hug.types.text,
        threshold: hug.types.float_number = 1.0,
        config: hug.types.text = None,
    ):
        config = get_config_from_file(config)
        bigsi = BIGSI(config)
        return bigsi.search(seq, threshold)

    @hug.object.cli
    @hug.object.delete("/", output_format=hug.output_format.json)
    def delete(self, config: hug.types.text = None):
        config = get_config_from_file(config)
        get_storage(config).delete_all()


def main():
    API.cli()


if __name__ == "__main__":
    main()



#! /usr/bin/env python
from __future__ import print_function
from bigsi.graph import BIGSI as Graph
import argparse
import json
import pickle


def delete_sample(sample_name, graph):
    if not sample_name:
        raise ValueError("Delete requires a sample name")
    try:
        graph.delete_sample(sample_name)
    except ValueError as e:
        return {"result": "failed", "error": "%s" % e}
    return {"result": "deleted %s" % sample_name}


def samples(sample_name, graph, delete=False):
    if delete:
        return delete_sample(sample_name, graph)
    else:
        if sample_name is None:
            out = {}
            for colour, sample_name in graph.colours_to_sample_dict().items():
                if not sample_name == "DELETED":
                    if not sample_name in out:
                        out[sample_name] = {}
                    if colour:
                        out[sample_name]["colour"] = int(colour)
                        out[sample_name]["name"] = sample_name
                        # out[sample_name]["kmer_count"] = graph.count_kmers(sample_name)
        else:
            out = {sample_name: {}}
            out[sample_name]["colour"] = graph.get_colour_from_sample(sample_name)
            out[sample_name]["name"] = sample_name
            # out[sample_name]["kmer_count"] = graph.count_kmers(sample_name)

        return out



from bigsi.storage.base import BaseStorage
from bigsi.constants import DEFAULT_BERKELEY_DB_STORAGE_CONFIG
import bsddb3
import os


class BerkeleyDBStorage(BaseStorage):
    def __init__(self, storage_config=None):
        if storage_config is None:
            storage_config = DEFAULT_BERKELEY_DB_STORAGE_CONFIG
        self.storage_config = storage_config
        self.storage = bsddb3.hashopen(
            storage_config["filename"],
            flag="c",
            cachesize=storage_config.get("hashsize", 20480),
        )

    def __repr__(self):
        return "berkeleydb Storage"

    def delete_all(self):
        self.storage.close()
        try:
            os.remove(self.storage_config["filename"])
        except FileNotFoundError:
            pass
        BerkeleyDBStorage.__init__(self, storage_config=self.storage_config)

    def sync(self):
        self.storage.sync()



#! /usr/bin/env python
from __future__ import print_function
from bigsi.graph import BIGSI as Graph
import os.path
import sys
import logging
import json

logger = logging.getLogger(__name__)
from bigsi.utils import DEFAULT_LOGGING_LEVEL

logger.setLevel(DEFAULT_LOGGING_LEVEL)
from bitarray import bitarray
from bigsi.cmds.build import load_bloomfilter


def insert(index, bloomfilter, sample):
    ## To do add warning that build is normally preferable
    index.insert(load_bloomfilter(bloomfilter), sample)
    return {"result": "success"}



# from distutils.core import setup
from setuptools import setup
import os


setup(
    name="bigsi",
    version="0.3.1",
    packages=[
        "bigsi",
        "bigsi.bloom",
        "bigsi.cmds",
        "bigsi.utils",
        "bigsi.graph",
        "bigsi.storage",
        "bigsi.matrix",
        "bigsi.scoring",
        "bigsi.tests",
    ],
    keywords="DBG coloured de bruijn graphs sequence search signture files signature index bitsliced",
    license="MIT",
    url="http://github.com/phelimb/bigsi",
    description="BItsliced Genomic Signature Index - Efficient indexing and search in very large collections of WGS data",
    author="Phelim Bradley",
    author_email="wave@phel.im",
    install_requires=[
        "cython",
        "hug",
        "numpy",
        "mmh3",
        "bitarray",
        "redis",
        "biopython",
        "pyyaml",
        "humanfriendly",
    ],
    entry_points={"console_scripts": ["bigsi = bigsi.__main__:main"]},
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
)



#! /usr/bin/env python
from __future__ import print_function

# from bigsi.utils import min_lexo
from bigsi.utils import seq_to_kmers
from bigsi.graph import BIGSI as Graph
import argparse
import os.path
import time
from Bio import SeqIO
import json
import logging
import sys

logger = logging.getLogger(__name__)
from bigsi.utils import DEFAULT_LOGGING_LEVEL

logger.setLevel(DEFAULT_LOGGING_LEVEL)
import operator
from bigsi.utils import convert_query_kmer


def per(i):
    return float(sum(i)) / len(i)


def parse_input(infile):
    gene_to_kmers = {}
    with open(infile, "r") as inf:
        for record in SeqIO.parse(inf, "fasta"):
            gene_to_kmers[record.id] = str(record.seq)
            yield (record.id, str(record.seq))
    # return gene_to_kmers


def _search(gene_name, seq, results, threshold, graph, output_format="json", pipe=False, score=False):
    if pipe:
        if output_format == "tsv":
            start = time.time()
            result = graph.search(seq, threshold=threshold, score=score)
            diff = time.time() - start
            if result:
                for sample_id, percent in result.items():
                    print(
                        "\t".join(
                            [gene_name, sample_id, str(round(percent["percent_kmers_found"], 2)), str(round(diff, 2))]
                        )
                    )
            else:
                print("\t".join([gene_name, "NA", str(0), str(diff)]))
        elif output_format == "fasta":
            samples = graph.sample_to_colour_lookup.keys()
            print(" ".join([">", gene_name]))
            print(seq)
            result = graph.search(seq, threshold=threshold, score=score)
            result = sorted(result.items(), key=operator.itemgetter(1), reverse=True)
            for sample, percent in result:
                percent = round(percent * 100, 2)
                colour = int(graph.sample_to_colour_lookup.get(sample))
                print(" ".join([">", gene_name, sample, "kmer-%i coverage %f" % (graph.kmer_size, percent)]))
                presence = []
                for kmer in seq_to_kmers(seq, graph.kmer_size):
                    kmer_presence = graph.graph.lookup(convert_query_kmer(kmer))[colour]
                    sys.stdout.write(str(int(kmer_presence)))
                sys.stdout.write("\n")
        else:
            result = {}
            start = time.time()
            result["results"] = graph.search(seq, threshold=threshold, score=score)
            diff = time.time() - start
            result["time"] = diff
            print(json.dumps({gene_name: result}))
    else:
        results[gene_name] = {}
        start = time.time()
        results[gene_name]["results"] = graph.search(seq, threshold=threshold, score=score)
        diff = time.time() - start
        results[gene_name]["time"] = diff
    return results


def search(seq, fasta_file, threshold, graph, output_format="json", pipe=False, score=False):
    if output_format == "tsv":
        print("\t".join(["gene_name", "sample_id", str("kmer_coverage_percent"), str("time")]))
    results = {}
    if fasta_file is not None:
        for gene, seq in parse_input(fasta_file):
            results = _search(
                gene_name=gene,
                seq=seq,
                results=results,
                threshold=threshold,
                graph=graph,
                output_format=output_format,
                pipe=pipe,
                score=score,
            )
    else:
        results = _search(
            gene_name=seq,
            seq=seq,
            results=results,
            threshold=threshold,
            graph=graph,
            output_format=output_format,
            pipe=pipe,
            score=score,
        )
    return results
